/*
** Simple C-like example of back propagation learning using
** matrix-matrix with logistic nonlinearities on 
** both the hidden units and output units.
** $Header: /da/bilmes/phipac/backprop/RCS/backprop.c,v 1.3 1995/12/14 01:52:24 bilmes Exp $
** Jeff Bilmes <bilmes@icsi.berkeley.edu>
*/


#include <stdlib.h>
#include <stdio.h>
#include <math.h>


#define BIAS_UNIT_VALUE 1.0

#define LOGISTIC(x) (1.0/(1.0+exp(-(x))))


extern void mul_mfmft_mf_beta0 (const int M, const int K, const int N, 
				const float *const A, 
				const float *const B, 
				float *const C, 
				const int Astride, 
				const int Bstride, 
				const int Cstride);

extern void mul_mfmf_mf_beta0 (const int M, const int K, const int N, 
			       const float *const A, 
			       const float *const B, 
			       float *const C, 
			       const int Astride, 
			       const int Bstride, 
			       const int Cstride);

extern void mul_mftmf_mf_alpha (const int M, const int K, const int N, 
				const float *const A, 
				const float *const B, 
				float *const C, 
				const int Astride, 
				const int Bstride, 
				const int Cstride, 
				const float alpha);

void
applysigmoid(float *mat,int M, int N, int stride)
{
  float *endp = &mat[M*stride];
  while (mat != endp) {
    float *rmat = mat;
    float *rendp = &mat[N];
    while (rmat != rendp) {
      *rmat = LOGISTIC(*rmat);
      rmat++;
    }
    mat += stride;
  }
}


void
applysoftmax(float *matrix,int M, int N, int stride)
{
  float *endp = &matrix[M*stride];
  float *row = matrix;
  while (row != endp) {
    float *rendp = &row[N];
    float *col = row;
    float sum = 0;
    while (col != rendp) {
      *col = exp(*col);
      sum += *col++; 
    }
    sum = 1.0/sum;
    col = row;
    while (col != rendp) {
      *col++ *= sum;
    }
    row += stride;
  }
}


float
sum_squared_error(float *opats,float *tpats,int Np, int No)
{
  float *endp = &opats[Np*No];
  float sum = 0.0;
  while (opats != endp ) {
    float tmp = (*tpats++ - *opats++);
    sum += tmp*tmp;
  }
  return sum;
}

void
compute_delta_outputs(float *delta_outputs,
		      float *output_patterns,
		      float *target_patterns,
		      int Np,
		      int No)
{
  float *endp = &delta_outputs[Np*No];
 
  while (delta_outputs != endp) {
    *delta_outputs = (*output_patterns)*(1-(*output_patterns))*
      ((*target_patterns)-(*output_patterns));
    delta_outputs++;
    output_patterns++;
    target_patterns++;
  }
}

void
finish_compute_delta_hiddens(
		      float *dhids,
		      float *hpats,
		      int M,
		      int N,
		      int hp_stride)
{
  float *endp = &hpats[M*hp_stride];
		     
  while (hpats != endp) {
    float *rendp = &hpats[N];
    float *hpatsp = hpats;
    while (hpatsp != rendp) {
      *dhids *= (*hpatsp)*(1-(*hpatsp));
      dhids ++ ;
      hpatsp ++ ;
    }
    hpats += hp_stride;
  }
}




float
bunched_mlp2_backprop (
       const int Ni,     /* Number input units */
       const int Nh,     /* Number hidden units */
       const int No,     /* Number output units */
       const int Np,     /* number patterns in a bunch (i.e., bunch size). */
       float *who, /* hidden to output weight matrix NoX(Nh+1) */
       float *wih, /* input to hidden weight matrix NhX(Ni+1) */
       /* for ipats and tpats, there are Np patterns (the bunch size)
	  and each one is a row in the corresponding matrix. */
       float *ipats, /* input patterns  NpX(Ni+1) */
       float *tpats, /* target patterns NpXNo */
       /* preallocated from the outside */
       float *hpats,  /* hidden layer NpX(Nh+1) */
       float *dhids,  /* delta hiddens NpXNh */
       float *opats,  /* output layer NpXNo */
       float *douts,  /* delta outputs NpXNo  */
       float alpha )  /* learning rate */
{
  
  /*******************/
  /* do forward pass */
  /*******************/
  
  /* Compute the hidden layer: 
     Use C = A*Transpose(B) so that the input and hidden patterns 
     can be stored as rows giving better cache performance. 
     */
  mul_mfmft_mf_beta0(Np,Ni+1,Nh,ipats,wih,hpats,Ni+1,Ni+1,Nh+1);
  applysigmoid(hpats,Np,Nh,Nh+1);

  /* Compute the output layer;
     Use C = A*Transpose(B) */
  mul_mfmft_mf_beta0(Np,Nh+1,No,hpats,who,opats,Nh+1,Nh+1,No);
  applysigmoid(opats,Np,No,No);

  /* we now have Np output vectors of length No */

  /****************************/
  /* propagate error backward */
  /****************************/

  compute_delta_outputs(douts,opats,tpats,Np,No);

  /* Partially compute delta_hiddens = delta_outputs*who
     Use C = A*B */
  mul_mfmf_mf_beta0(Np,No,Nh,douts,who,dhids,
		    No,Nh+1,Nh);
  finish_compute_delta_hiddens(dhids,hpats,Np,Nh,Nh+1);


  /* Adjust hidden-output weights:
     Use C = alpha*transpose(A)*B + C */
  mul_mftmf_mf_alpha(No,Np,Nh+1,douts,hpats,who,
		     No,Nh+1,Nh+1,alpha);

  /* Adjust input-hidden weights:
     Use C = alpha*transpose(A)*B + C */
  mul_mftmf_mf_alpha(Nh,Np,Ni+1,dhids,ipats,wih,
		     Nh,Ni+1,Ni+1,alpha);

  /* return the sum-squared error */
  return sum_squared_error(opats,tpats,Np,No);

}




void
bunched_mlp2_forwprop (
	  int Ni,     /* Number input units */
	  int Nh,     /* Number hidden units */
	  int No,     /* Number output units */
	  int Np,     /* number patterns in a bunch (i.e., bunch size). */
	  float *who, /* hidden to output weight matrix NoX(Nh+1) */
	  float *wih, /* input to hidden weight matrix NhX(Ni+1) */
	  /* for ipats and tpats, there are Np patterns (the bunch size)
	     and each one is a row in the corresponding matrix. */
	  float *ipats, /* input patterns  NpX(Ni+1) */
	  /* preallocated from the outside */
	  float *hpats,  /* hidden layer NpX(Nh+1) */
	  float *opats   /* output layer NpXNo */
	  ) 
{
  
  /*******************/
  /* do forward pass */
  /*******************/
  
  /* Compute the hidden layer: 
     Use C = A*Transpose(B) so that the input and hidden patterns 
     can be stored as rows giving better cache performance. 
     */
  mul_mfmft_mf_beta0(Np,Ni+1,Nh,ipats,wih,hpats,Ni+1,Ni+1,Nh+1);
  applysigmoid(hpats,Np,Nh,Nh+1);

  /* Compute the output layer;
     Use C = A*Transpose(B) */
  mul_mfmft_mf_beta0(Np,Nh+1,No,hpats,who,opats,Nh+1,Nh+1,No);
  applysigmoid(opats,Np,No,Nh+1);

  /* we now have Np output vectors of length No */

}
